import os
from typing import Optional
import configparser

from loguru import logger


__all__ = [
    'Config',
    'CONFIG_BASE_PATH',
    'CONFIG_FILE_NAME'
]


CONFIG_BASE_PATH = os.path.expanduser('~/.config/sen-api')
CONFIG_FILE_NAME = 'config.ini'


class Config(object):
    def __init__(self, base_path=CONFIG_BASE_PATH, config_file_name=CONFIG_FILE_NAME):
        if not os.path.isdir(base_path):
            os.mkdir(base_path)
        self.base_path = base_path
        self.path = os.path.join(base_path, config_file_name)
        self._config = configparser.ConfigParser()

    def load(self) -> bool:
        """
        :return: True if config loaded successfully
        """
        if not os.path.isfile(self.path):
            logger.debug('Configuration file not found, skipping')
            return False
        self._config = configparser.ConfigParser()
        with open(self.path, 'r') as f:
            logger.debug('Reading configuration file...')
            self._config.read_file(f)
        return True

    def write(self, section: str, values: dict):
        logger.debug('Writing config...')
        for key in values.keys():
            if not self._config.has_section(section):
                self._config.add_section(section)
            self._config[section][key] = values[key]
        with open(self.path, 'w') as f:
            self._config.write(f)

    def get_value(self, section: str, value: str, fallback=None) -> Optional:
        return self._config.get(section, value, fallback=fallback)



import os
import pickle
import requests
from typing import Optional, List

from bs4 import BeautifulSoup
from loguru import logger

from sen_api import IntervalReading, Config, Bill, AuthenticationError


__all__ = [
    'SENProvider'
]


class SENProvider(object):
    _base_url = 'https://www.servizioelettriconazionale.it/it-IT'
    _meter_url = f'{_base_url}/clienti/SEN/servizi/Areaclienti/Contatore/a.ser?tab=3'
    _client_area_url = f'{_base_url}/clienti/SEN/servizi/Areaclienti/HomePage/homepage.jsp'
    _bills_url = f'{_base_url}/clienti/SEN/servizi/Areaclienti/SelezionaBolletteServlet/a.ser'
    _bill_download_url = f'{_base_url}/clienti/SEN/servizi/Areaclienti/DettaglioBolletta/vediPDF.ser?from=bollettaPDF'
    _meter_readings_url = f'{_base_url}/clienti/SEN/servizi/Areaclienti/LeggiConsumi/a.ser?funz=A09&destMenu=areaclienti_left.jsp&from=modifica'

    def __init__(self, config: Config):
        self._session = requests.Session()
        self._session_path = os.path.join(config.base_path, 'session.pickle')
        self._config = config
        self._client_id = None
        self._client_name = None

    @staticmethod
    def _get_soup(data) -> BeautifulSoup:
        return BeautifulSoup(data, 'html.parser')

    def _send_form(self, form=None, soup_data: Optional[str] = None, form_data: Optional[dict] = None):
        if not form:
            soup = self._get_soup(soup_data)
            if not form_data:
                form_data = dict()
            form = soup.find('form')

        action_url = form.get('action')
        for field in form.find_all('input'):
            name = field.get('name')
            value = field.get('value', None)
            if name not in form_data.keys():
                form_data[name] = value

        response = self._session.post(action_url, data=form_data, allow_redirects=True)
        return response

    def _real_auth(self, username, password):
        self._session = requests.Session()

        logger.debug('Getting base url...')
        response = self._session.get(self._base_url)

        # first login form
        login_data = {
            'txtUsername': username,
            'txtPassword': password
        }
        login_response = self._send_form(soup_data=response.text, form_data=login_data)
        logger.debug('Got login response')

        # saml request
        response = self._send_form(soup_data=login_response.text)
        logger.debug('Done saml request')

        # saml response
        response = self._send_form(soup_data=response.text)
        logger.debug('Got saml response')

        soup = self._get_soup(response.text)
        self._client_name = soup.find('h3', id='nomeCliente').text.strip()
        self._client_id = soup.find('a', id='tabsForniture_selezionata').find('b').text
        logger.debug(f'Client name is: {self._client_name}')
        logger.debug(f'Client ID is: {self._client_id}')
        self._config.write(section='client', values={'name': self._client_name, 'id': self._client_id})

    @property
    def is_authenticated(self) -> bool:
        """
        Check if it is still authenticated
        """
        if not self.load_session():
            return False
        if not self._session.cookies:
            return False
        logger.debug('Checking if session is still valid...')
        response = self._session.get(self._client_area_url)
        return response.url == self._client_area_url

    @property
    def client_info(self) -> dict:
        return {
            'id': self._client_id if self._client_id else self._config.get_value('client', 'id'),
            'name': self._client_name if self._client_name else self._config.get_value('client', 'name')
        }

    def save_session(self):
        logger.debug('Saving session...')
        with open(self._session_path, 'wb') as f:
            pickle.dump(self._session, f)

    def load_session(self) -> bool:
        """
        :return: True if loaded, False otherwise
        """
        logger.debug('Loading session...')
        if os.path.isfile(self._session_path):
            with open(self._session_path, 'rb') as f:
                self._session = pickle.load(f)
                return True
        return False

    def authenticate(self, username: Optional[str] = None, password: Optional[str] = None, force=False):
        if not username:
            username = self._config.get_value('auth', 'username')
        if not password:
            password = self._config.get_value('auth', 'password')

        if username is None or password is None:
            raise ValueError('Credentials cannot be None.')

        if not force:
            self.load_session()
        if force or not self.is_authenticated:
            try:
                self._real_auth(username, password)
                self.save_session()
            except AttributeError:
                message = 'Authentication error or wrong credentials.'
                logger.error(message)
                raise AuthenticationError(message)
        logger.debug('Successfuly authenticated.')

    def get_last_reading(self) -> dict:
        response = self._session.get(self._meter_url)
        soup = self._get_soup(response.text)
        table = soup.find('table', attrs={'class': 'pe_tabsData tabella_contatore'})
        cells = table.find_all('td')
        reading_date = cells[7].text
        readings = cells[5].text.split()
        return {
            'reading_date': reading_date,
            'readings': {
                'A1': readings[1],
                'A2': readings[3],
                'A3': readings[5]
            }
        }

    def get_all_readings(self) -> List[IntervalReading]:
        response = self._session.get(self._meter_readings_url)
        soup = self._get_soup(response.text)
        table = soup.find('table', id='tabella_consumi')
        readings = []
        for row in table.find_all('tr', attrs={'class': 'border border-right'}):
            cells = row.find_all('td')
            r = IntervalReading(
                interval_start=cells[0].text,
                interval_end=cells[1].text,
                total_consumption=int(cells[3].text)
            )
            readings.append(r)
        return readings

    def get_bills_available_years(self) -> List[str]:
        response = self._session.post(self._bills_url)
        soup = self._get_soup(response.text)
        years = soup.find('div', id='sceltaanni').find_all('a')
        return [y.text for y in years]

    def get_bills(self, year: str) -> List[Bill]:
        if year not in self.get_bills_available_years():
            error = f'Year {year} is not available'
            logger.debug(error)
            raise ValueError(error)
        data = {'annoScelto': year}
        response = self._session.post(self._bills_url, data=data)
        soup = self._get_soup(response.text)
        table = soup.find('table', id='tab_bollette')
        bills = []
        for row in table.find_all('tr', attrs={'class': year}):
            # store hidden inputs in params dict, used later to download the bill
            params_input = row.find_all('input')
            params = dict()
            for p in params_input:
                key = p.get('name')
                value = p.get('value')
                params[key] = value

            # get '1' from 'codFatt_1', '2' from 'codFatt_2'...
            params['occorrenzaForm'] = list(params.keys())[0].split('_')[1]

            cells = row.find_all('td')
            bill = Bill(
                number=int(cells[0].text),
                due_date=cells[1].text,
                amount=float(cells[2].text.replace(',', '.')),
                is_payed=(cells[5].text.strip() == 'Incassata totalmente'),
                params=params
            )
            bills.append(bill)

        return bills

    def download_bill(self, bill: Bill, download_path: Optional[str] = None) -> Optional[str]:
        data = {
            'tipoRichiesta': '2',
        }
        data.update(bill.params)
        response = self._session.post(self._bill_download_url, data=data)
        if response.headers['content-type'] == 'application/pdf':
            path = os.path.join(self._config.base_path, 'bills')
            if download_path:
                path = download_path
            if not os.path.isdir(path):
                os.mkdir(path)
            path = os.path.join(path, bill.document_name)
            with open(path, 'wb') as f:
                f.write(response.content)
            return path
        return None



import math
from functools import wraps
from json import dumps as json_dumps

import click
from click import echo, clear
from halo import Halo
from rich.console import Console
from rich.table import Table
from loguru import logger

from sen_api import SENProvider, Config, __version__, AuthenticationError


__all__ = [
    'cli'
]


config = Config()
provider = SENProvider(config=config)

console = Console()


def auth_required(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        # if not already authenticated, try to authenticate
        if not provider.is_authenticated:
            try:
                provider.authenticate()
            except (AuthenticationError, ValueError):
                echo('Cannot authenticate.')
                return None
        return f(*args, **kwargs)
    return wrapper

########################################################################################################################


@click.group()
@click.version_option(__version__)
@click.option('--verbose', '-v', help='Enable verbose logs.', is_flag=True)
@click.option('--json', '-j', help='Print in JSON format when possible.', is_flag=True)
@click.pass_context
def cli(ctx, verbose, json):
    ctx.ensure_object(dict)
    if not verbose:
        logger.remove()
    ctx.obj['JSON'] = json
    config.load()

########################################################################################################################


def _authenticate(username=None, password=None, force=False):
    clear()

    success = True
    with Halo(text='Authenticating', spinner='dots'):
        try:
            if force:
                provider.authenticate(force=True)
            else:
                provider.authenticate(username, password, force=True)
        except (AuthenticationError, ValueError):
            logger.error('Error while trying to authenticate.')
            success = False

    if success:
        if not force:
            config.write('auth', {
                'username': username,
                'password': password
            })
        echo('Successfully authenticated.')
    else:
        echo('Cannot authenticate, check your credentials.')


def force_authentication(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    _authenticate(force=True)
    ctx.exit()


@cli.command()
@click.option('--username', '-u', prompt=True)
@click.option('--password', '-p', prompt=True, hide_input=True, confirmation_prompt=True)
@click.option('--force', '-f', help='Force authentication using saved credentials.', is_flag=True,
              callback=force_authentication, expose_value=False, is_eager=True)
def authenticate(username, password):
    _authenticate(username, password)

########################################################################################################################


@auth_required
def last_reading(json):
    reading = provider.get_last_reading()

    if json:
        echo(json_dumps(reading))
    else:
        band_readings = reading['readings']

        table = Table(title=reading['reading_date'])
        table.add_column('Time band')
        table.add_column('Consumption')

        table.add_row('A1', band_readings['A1'], style='red')
        table.add_row('A2', band_readings['A2'], style='yellow')
        table.add_row('A3', band_readings['A3'], style='green')

        console.print(table)


@auth_required
def all_readings(json):
    readings_list = provider.get_all_readings()

    if json:
        echo(json_dumps([r.to_dict() for r in readings_list]))
    else:
        avg_consumption_avg = 0
        if len(readings_list) != 0:
            avg_consumption_avg = math.ceil(sum(r.avg_consumption for r in readings_list) / len(readings_list))

        table = Table(title='Readings')
        table.add_column('Start', style='cyan')
        table.add_column('End', style='green')
        table.add_column('Days')
        table.add_column('Total consumption', justify='center')
        table.add_column('Mean consumption', justify='center')
        for r in readings_list:
            color = 'red' if r.avg_consumption > avg_consumption_avg else 'default'
            table.add_row(
                str(r.interval_start.date()),
                str(r.interval_end.date()),
                str(r.interval_days),
                str(r.total_consumption),
                f'[{color}]{str(r.avg_consumption)}[/{color}]'
            )

        console.print(table)
        console.print('Values above the mean are colored [red]red[/red].\n')


@cli.command()
@click.option('--all', '-a', '_all', help='Get all readings.', is_flag=True)
@click.option('--last', '-l', help='Get only the last readings.', is_flag=True)
@click.pass_context
def readings(ctx, _all, last):
    json = ctx.obj['JSON']
    if _all:
        all_readings(json)
    elif last:
        last_reading(json)
    else:
        echo(ctx.get_help())
        ctx.exit()

########################################################################################################################


@cli.command()
@click.pass_context
@auth_required
def client_info(ctx):
    values = provider.client_info

    if ctx.obj['JSON']:
        echo(json_dumps(values))
    else:
        table = Table(title='Client info')
        table.add_column('Field')
        table.add_column('Value')
        for v in values.keys():
            table.add_row(v, values[v])

        console.print(table)

########################################################################################################################


@cli.command()
@click.option('--year', '-y', help='Specify the bills year.')
@click.option('--download', '-d', type=int, help='Download bill with the specified in PDF format.')
@click.pass_context
@auth_required
def bills(ctx, year, download):
    json = ctx.obj['JSON']
    if not year:
        years = provider.get_bills_available_years()
        if json:
            echo(json_dumps({'available_years': years}))
        else:
            table = Table(title='Available years')
            table.add_column('Year', width=15, justify='center')
            for year in years:
                table.add_row(year)

            console.print(table)
    else:
        try:
            bills_list = provider.get_bills(year=year)
        except ValueError as e:
            echo(e)
            ctx.exit()

        if download:
            found_bill = None
            for bill in bills_list:
                if bill.number == download:
                    found_bill = bill

            if not found_bill:
                echo(f'Bill with number {download} not found.')
                ctx.exit()

            with Halo(text=f'Downloading {found_bill.document_name} ...', spinner='dots'):
                download_path = provider.download_bill(found_bill)

            if download_path:
                console.print(f'Bill successfuly downloaded in [cyan]{download_path}[/cyan]')
            else:
                echo('Error during the download, enable verbose output for more details.')

        else:
            if json:
                echo(json_dumps([b.to_dict() for b in bills_list]))
            else:
                avg_amount = 0
                if len(bills_list) != 0:
                    avg_amount = math.ceil(sum(b.amount for b in bills_list) / len(bills_list))

                table = Table(title=f'Year [cyan]{year}[/cyan] bills')
                table.add_column('Number')
                table.add_column('Due date')
                table.add_column('Amount')
                table.add_column('Payed')
                for b in bills_list:
                    amount_color = 'red' if b.amount > avg_amount else 'default'
                    table.add_row(
                        str(b.number),
                        str(b.due_date.date()),
                        f'[{amount_color}]{str(b.amount)}€[/{amount_color}]',
                        '[green]Yes[/green]' if b.is_payed else '[red]No[/red]'
                    )

                console.print(table)
                console.print('Values above the mean are colored [red]red[/red].\n')



import math

from .utils import str_to_datetime


__all__ = [
    'IntervalReading',
    'Bill'
]


class IntervalReading(object):
    def __init__(self, interval_start: str, interval_end: str, total_consumption: int):
        self.total_consumption = total_consumption
        self.interval_start = str_to_datetime(interval_start)
        self.interval_end = str_to_datetime(interval_end)
        self.interval_days = (self.interval_end - self.interval_start).days + 1   # count the last day too
        self.avg_consumption = 0 if self.interval_days == 0 else math.ceil(self.total_consumption / self.interval_days)

    def to_dict(self) -> dict:
        return {
            'interval_start': str(self.interval_start.date()),
            'interval_end': str(self.interval_end.date()),
            'interval_days': self.interval_days,
            'total_consumption': self.total_consumption,
            'mean_consumption': self.avg_consumption
        }

    def __eq__(self, other: 'IntervalReading'):
        return (
            self.interval_start == other.interval_start,
            self.interval_end == other.interval_end,
            self.total_consumption == other.total_consumption
        )

    def __str__(self):
        return f'<IntervalReading ' \
               f'start={str(self.interval_start.date())},' \
               f'end={self.interval_end.date()},' \
               f'consumption={self.total_consumption}>'

    def __repr__(self):
        return self.__str__()

########################################################################################################################


class Bill(object):
    def __init__(self, number: int, due_date: str, amount: float, is_payed: bool, params: dict):
        self.number = number
        self.due_date = str_to_datetime(due_date)
        self.amount = amount
        self.is_payed = is_payed
        self.params = params

    @property
    def document_name(self) -> str:
        return f'bill_{str(self.due_date.date())}.pdf'

    def to_dict(self) -> dict:
        return {
            'number': self.number,
            'due_date': str(self.due_date.date()),
            'amount': self.amount,
            'is_payed': self.is_payed
        }

    def __str__(self):
        return f'<Bill ' \
               f'number={str(self.number)},' \
               f'due_date={str(self.due_date.date())},' \
               f'amount={str(self.amount)},' \
               f'is_payed={str(self.is_payed)}>'

    def __repr__(self):
        return self.__str__()



from setuptools import setup, find_packages

from sen_api import __version__, __author__


with open('README.md', 'r') as f:
    long_description = f.read()


setup(
    name='sen-api',
    version=__version__,
    author=__author__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['tests*']),
    install_requires=[
        'requests',
        'beautifulsoup4',
        'click',
        'halo',
        'loguru',
        'rich'
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'sen-api = sen_api.__main__:main'
        ]
    }
)
